/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/database/manager.c
 * @author Christian Grothoff
 *
 * This module is responsible to manage content, in particular 
 * it needs to decide when to keep content and when to actively
 * push content out into the network.
 **/

#include "config.h"
#include <stdio.h>
#include <dirent.h>
#include <errno.h>
#include "util/hashing.h"
#include "util/checksum.h"
#include "util/semaphore.h"
#include "util/statistics.h"
#include "util/logging.h"
#include "applications/afs/database/manager.h"
#include "applications/afs/database/fileindex.h"
#include "applications/afs/database/contentdatabase.h"
#include "applications/afs/util/contentencoding.h"
#include "applications/afs/encoding/block.h"
#include "applications/afs/protocol/afs.h"
#include "applications/afs/protocol/messages.h"

/* ********************* GLOBALS ***************** */

/**
 * What is the priority of the least important content
 * that we are storing? The scanning cron job updates
 * this value to indicate how low the priority of content
 * is that is currently being removed.
 *
 * The value is MANAGER_age if we have excess storage
 * space, allowing the node to store any type of content.
 **/
static double lowImportance;

/**
 * The current base value for fresh content (used to time-out old
 * content).
 **/
static int MANAGER_age;

/**
 * Is active migration allowed? This is about us receiving
 * data from the network, actively pushing content out is
 * always ok.
 **/
static int useActiveMigration;

/**
 * Handle to the thread discarding content.
 **/
static LookupDatabaseIterator * ldi_discarding;

/**
 * LookupDatabaseIterator for content migration
 **/
static LookupDatabaseIterator * ldi_migration;

/**
 * Global database handle
 **/
static DatabaseHandle * dbHandle = NULL;

/**
 * Lock for ldi_migration
 **/
static Mutex lock;

/**
 * Lock for driveFullCallback
 **/
static Mutex driveFullLock;

/**
 * Shutdown flag.
 **/
static int manager_shutdown = YES;

/**
 * Statistics handles
 **/
static int stat_handle_lookup_3hash;
static int stat_handle_lookup_chk;
static int stat_handle_lookup_ondemand;
static int stat_handle_drivefull;
static int stat_handle_content_pushed;

/**
 * Callback for statuscalls when drive full.
 **/
static void driveFullCallback(long long * spaceleft) {
  ContentIndex entry;
  CollisionLevel * level;
  double out;

  MUTEX_LOCK(&driveFullLock);
  LOG(LOG_DEBUG,
      "DEBUG: drive full, freeing %d kb\n",
      *spaceleft);
  statChange(stat_handle_drivefull, 1);
  out = 0.0;
  while ((*spaceleft < 0) && (manager_shutdown == NO)) {
    if (NO == ldiHasNext(ldi_discarding)) {
      destroyLDI(ldi_discarding);
      ldi_discarding = createLDI();
      if (NO == ldiHasNext(ldi_discarding))
	return; /* nothing can be done */
    }
    ldiGetNext(ldi_discarding,
	       &entry,
	       &level);
    switch (ntohs(entry.type)) {
    case LOOKUP_TYPE_FREE:
    case LOOKUP_TYPE_DELETED:
      continue;
    case LOOKUP_TYPE_CHK:
    case LOOKUP_TYPE_3HASH:
    case LOOKUP_TYPE_SUPER:
      if (ntohs(entry.fileNameIndex) > 0)
	continue;
      else 
	break;
    default:
      LOG(LOG_WARNING,
	  "WARNING: encountered unexected content type %d, "\
	  "discard thread aborts (run gnunet-check -a to fix)\n",
	  ntohs(entry.type));
      return;
    }
    if (lowImportance < 1.0)
      lowImportance = 1.0;
    if (ntohl(entry.importance) <= lowImportance) {
      /* FIXME: we may want to give the content a chance to 
	 migrate elsewhere... */
      LOG(LOG_DEBUG,
	  "DEBUG: freeing content that had priority %u <= %lf\n",
	  ntohl(entry.importance), 
	  lowImportance);
      delEntry(&entry, level);
      unlinkFromDB(dbHandle, &entry.hash);
      (*spaceleft)++;
      /* back off a bit (10 times as fast as usual should be ok) */
      lowImportance = (lowImportance * 9.0)/10.0;
    } else {
      LOG(LOG_DEBUG,
	  "DEBUG: keeping content that had priority %u > %f, searching for more...\n",
	  ntohl(entry.importance), 
	  lowImportance);
      out += 1.0;
      lowImportance = out + (lowImportance * 99.0 + (double)ntohl(entry.importance) + (10.0*(*spaceleft)))/99.0; 
      /* + *spaceleft% to do something faster if we need to discard lots... */
    }
  }
  MUTEX_UNLOCK(&driveFullLock);
  LOG(LOG_DEBUG,
      "DEBUG: drive full processing done.\n");
}

/**
 * Open the AGE file and return the handle.
 **/
static int getAgeFileHandle() {
  char * fileName;
  int handle;
  
  LOG(LOG_CRON, 
      "CRON: enter cronReduceImportance\n");
  fileName = getFileName("AFS",
			 "AGEFILE",
			 "Configuration file must specify filename for"\
			 " the age of the database in section %s under %s.\n");
  handle = open(fileName,
		O_CREAT|O_RDWR,
		S_IRUSR|S_IWUSR);
  if (handle < 0) {
    LOG(LOG_WARNING,
	"Could not open agefile %s (%s)\n",
	fileName,
	strerror(errno));
    xfree(fileName,
	  "ageFile freeing");
    return SYSERR;
  }
  xfree(fileName,
	"ageFile freeing");
  return handle;
}

/**
 * Cron-job that decreases the importance-level of all
 * files by 1. Runs 'not very often'.
 **/
static void cronReduceImportance(void * unused) {
  int handle;
    
  LOG(LOG_CRON, 
      "CRON: enter cronReduceImportance\n");
  handle = getAgeFileHandle();
  if (handle == SYSERR)
    return;
  MANAGER_age++;
  write(handle, 
	&MANAGER_age,
	sizeof(int));
  close(handle);
  LOG(LOG_CRON, 
      "CRON: exit cronReduceImportance\n");
}

/**
 * Encode a block from a file on the drive, put the
 * result in the result buffer (allocate) and return
 * the size of the block.
 **/ 
static int encodeOnDemand(ContentIndex * ce,
			  CONTENT_Block ** result) {
  char * fn;
  int fileHandle;
  ssize_t blen;
  HashCode160 hc;
  CONTENT_Block * iobuf;
  HexName hex;

  /* on-demand encoding mechanism */
  fn = getIndexedFileName(ntohs(ce->fileNameIndex));
  if (fn == NULL) {
    LOG(LOG_FAILURE, 
	"FAILURE: Database inconsistent! "\
	"(index points to invalid offset (%d)\n",
	ntohs(ce->fileNameIndex));
    return SYSERR;
  }
  fileHandle = open(fn, O_EXCL, S_IRUSR);
  if (fileHandle == -1) {
    LOG(LOG_FAILURE, 
	"FAILURE: Could not open file %s.\n",
	fn);   
    xfree(fn, "filename");
    return SYSERR;
  }
  xfree(fn, "filename");
  lseek(fileHandle, 
	ntohl(ce->fileOffset), 
	SEEK_SET);
  iobuf = xmalloc(sizeof(CONTENT_Block),
		  "allocating space for io buffer");
  blen = read(fileHandle, 
	      iobuf,
	      sizeof(CONTENT_Block));
  memset(&((char*)iobuf)[blen],
	 0,
	 sizeof(CONTENT_Block)-blen);
  /*LOG(LOG_DEBUG,
      "DEBUG: read %u bytes for on-demand encoding at %u\n",
      blen, 
      ntohl(ce->fileOffset));*/
  close(fileHandle);
  if (blen < 0) {
    LOG(LOG_ERROR,
	"ERROR: could not read file (%s)\n",
	strerror(errno));
    xfree(iobuf, 
	  "freeing io buffer");
    return SYSERR;
  }
  hash(iobuf,
       blen, 
       &hc);
  *result = xmalloc(sizeof(CONTENT_Block), 
		    "allocating buffer for on-demand encoding result");
  if (SYSERR == encryptContent(iobuf,
			       &hc,
			       *result)) 
    errexit("ERROR: encryption of content failed\n");

  xfree(iobuf, 
	"freeing io buffer");
  IFLOG(LOG_DEBUG,
	hash(*result,
	     sizeof(CONTENT_Block),
	     &hc);
	hash2hex(&hc,
		 &hex));
  /*LOG(LOG_DEBUG,
      "DEBUG: on-demand encoded content has query %s\n",
      &hex);*/
  return sizeof(CONTENT_Block);
}

/**
 * Select content for active migration.
 **/
static int selectMigrationContent(HostIdentity * receiver,
				  ContentIndex * ce) {
  unsigned int avDist;
  unsigned int dist;
  CollisionLevel * level;
  int itc;  
  cron_t startTime;
  
  cronTime(&startTime);
  itc = 0;
  avDist = 1;
  while (itc++ < 512) {
    if ( (cronTime(NULL) - startTime > 100 * cronMILLIS) ||
	 (getCPULoad() > 80) )
      return SYSERR; /* abort after 100ms or if load > 80 */
    if (NO == ldiHasNext(ldi_migration)) {
      destroyLDI(ldi_migration);
      ldi_migration = createLDI();
      if (NO == ldiHasNext(ldi_migration))
	return SYSERR; /* no content in DB! */
    }
    ldiGetNext(ldi_migration,
	       ce,
	       &level);
    dist = distanceHashCode160(&ce->hash,
			       &receiver->hashPubKey);
    if (randomi(dist) < randomi(avDist)) /* select */
      return OK;
    else
      avDist = (avDist * 31 + dist)/32;
  }
  return SYSERR;
}
				  
/**
 * Build a CHK reply message for some content
 * selected for migration.
 * @return OK on success, SYSERR on error
 **/
static int buildCHKReply(ContentIndex * ce,
			 AFS_p2p_CHK_RESULT * pmsg) {
  CONTENT_Block * data;

  if (ntohs(ce->type) != LOOKUP_TYPE_CHK)
    return SYSERR;
  data = NULL;
  if (SYSERR == retrieveContent(ce,
				&data)) {
    freeNonNull(data, 
		"unexpected reply from retrieveContent");
    return SYSERR;
  }
  pmsg->header.size 
    = htons(sizeof(AFS_p2p_CHK_RESULT));
  pmsg->header.requestType
    = htons(AFS_p2p_PROTO_CHK_RESULT);
  memcpy(&pmsg->result,
	 data,
	 sizeof(CONTENT_Block));
  xfree(data,
	"freeing CHK block");
  return OK;
}

/**
 * Callback method for pushing content into the network.
 * The method chooses either a "recently" deleted block
 * or content that has a hash close to the receiver ID
 * (randomized to guarantee diversity, unpredictability
 * etc.).<p>
 *
 * @param receiver the receiver of the message
 * @param position is the reference to the
 *        first unused position in the buffer where GNUnet is building
 *        the message
 * @param padding is the number of bytes left in that buffer.
 * @return the number of bytes written to
 *   that buffer (must be a positive number).
 **/
static int activeMigrationCallback(HostIdentity * receiver,
				   char * position,
				   int padding) {
  AFS_p2p_CHK_RESULT * pmsg;
  int res;
  void * data;
  ContentIndex ce;
  
  res = 0;
  MUTEX_LOCK(&lock);
  while (padding - res > sizeof(AFS_p2p_CHK_RESULT)) {
    data = NULL;
    if (SYSERR == selectMigrationContent(receiver,
					 &ce)) {
      MUTEX_UNLOCK(&lock);
      return res; /* nothing selected, that's the end */
    }
    /* append it! */
    pmsg = (AFS_p2p_CHK_RESULT*) &position[res];
    if (OK == buildCHKReply(&ce,
			    pmsg)) {
      statChange(stat_handle_content_pushed, 1);
      res += sizeof(AFS_p2p_CHK_RESULT);
    } else {
      MUTEX_UNLOCK(&lock);
      return res; /* abort early after any error */
    }
  }
  MUTEX_UNLOCK(&lock);
  return res;
}

/**
 * Initialize the manager-module.
 **/
void initManager() {  
  char * dtype;
  int dbuckets;
  int handle;

 
  dtype = getConfigurationString("AFS",
  	 	 	         "DATABASETYPE");
  if(!dtype)
    errexit("AFS/DATABASETYPE not specified in config\n");
  dbuckets = getConfigurationInt("AFS",
  				 "DATABASEBUCKETS");
  if(dbuckets <= 0)
    errexit("AFS/DATABASEBUCKETS not specified in config\n");
  dbHandle = initContentDatabase(dtype, dbuckets);
  xfree(dtype, "freeing dtype");

  manager_shutdown = NO;
  MUTEX_CREATE(&lock);
  stat_handle_lookup_3hash
    = statHandle("# lookup (3HASH, search results)");
  stat_handle_lookup_chk
    = statHandle("# lookup (CHK, inserted or migrated content)");
  stat_handle_lookup_ondemand
    = statHandle("# lookup (ONDEMAND, indexed content)");
  stat_handle_drivefull
    = statHandle("# times storage full");
  stat_handle_content_pushed
    = statHandle("# kb content pushed out as padding");
  handle = getAgeFileHandle();
  MANAGER_age = 0;
  if (handle != SYSERR) {
    read(handle, 
	 &MANAGER_age,
	 sizeof(int));
    close(handle);
  }
  useActiveMigration
    = testConfigurationString("AFS",
			      "ACTIVEMIGRATION",
			      "YES");
  if (YES == useActiveMigration) {
    ldi_discarding = createLDI();
    MUTEX_CREATE(&driveFullLock);
    registerDriveFullCallback(&driveFullCallback);    
    lowImportance = 1.0; 
  }
  ldi_migration = createLDI();
  coreAPI->registerSendCallback(sizeof(AFS_p2p_CHK_RESULT),
				(BufferFillCallback)&activeMigrationCallback);
  addCronJob(&cronReduceImportance,
	     6 * cronHOURS,
	     12 * cronHOURS,
	     NULL); 
}

/**
 * Shutdown the manager module.
 **/
void doneManager() {
  manager_shutdown = YES;
  coreAPI->unregisterSendCallback(sizeof(AFS_p2p_CHK_RESULT),
				  (BufferFillCallback)&activeMigrationCallback);
  delCronJob(&cronReduceImportance,
	     12 * cronHOURS,
	     NULL); 
  if (YES == useActiveMigration) {
    MUTEX_LOCK(&driveFullLock);
    unregisterDriveFullCallback(&driveFullCallback); 
    MUTEX_UNLOCK(&driveFullLock);
    MUTEX_DESTROY(&driveFullLock);
    destroyLDI(ldi_discarding);
  }
  destroyLDI(ldi_migration);
  MUTEX_DESTROY(&lock);
  doneContentDatabase(dbHandle);
}

/**
 * Locate content. This method locates the data matching the
 * content entry. The data is on-demand encrypted if it is
 * indexed content or retrieved from the contentdatabase
 * if it was inserted content. 
 *
 * @param ce the content entry describing what to look for
 * @param result where to write the result, space will be
 *        allocated by retrieveContent, *result should be
 *        NULL when this function is invoked.
 * @return the length of the resulting content, SYSERR on error
 **/
int retrieveContent(ContentIndex * ce,
		    CONTENT_Block ** result) {
  switch (ntohs(ce->type)) {
  case LOOKUP_TYPE_FREE:    
  case LOOKUP_TYPE_DELETED:
  case LOOKUP_TYPE_SUPER:
    return SYSERR;
  case LOOKUP_TYPE_CHK:
    if (ntohs(ce->fileNameIndex) == 0) {
      statChange(stat_handle_lookup_chk, 1);
      return readContent(dbHandle,
      			 &ce->hash,
			 (void**)result);
    } else {
      statChange(stat_handle_lookup_ondemand, 1);
      return encodeOnDemand(ce, result);
    }
  case LOOKUP_TYPE_3HASH:
    statChange(stat_handle_lookup_3hash, 1);
    return readContent(dbHandle,
    		       &ce->hash,
		       (void**)result);	 
  default:
    LOG(LOG_WARNING,
	"WARNING: unknown CE type %d in retrieveContent\n",
	ntohs(ce->type));
  }
  return SYSERR;  
}
 
/**
 * Content has been deemed valueable, the manager should take
 * this into account.
 * @param ce the content that was found to be valueable
 * @param deltaValue just how valueable was the content?
 **/
void contentValueable(ContentIndex * ce,
		      CollisionLevel * level,
		      int deltaValue) {
  ce->importance = htonl(deltaValue + ntohl(ce->importance));
  addEntry(ce, level);
}

/**
 * Store content (if the priority is high enough), potentially
 * discarding less important content.  Note that this method is never
 * called for indexed content since it is not considered to take any
 * space. If the manager decides to keep the content, it must call
 * addEntry on the lookup database, update the bloom filter and add
 * the data to the contentdatabase.
 * 
 * @param ce the content entry describing the content
 * @param len the length of the data in bytes
 * @param data the block itself
 * @param sender from where does the content come? NULL for
 *        from local client.
 * @param duplicate output param, will be YES if content was already there
 * @return OK if the block was stored, SYSERR if not
 **/
int insertContent(ContentIndex * ce,
		  int len,
		  void * data,
		  HostIdentity * sender,
		  int * duplicate) {
  void * old;
  int oldLen;
  
  *duplicate = NO;
  if ( (sender != NULL) &&
       (useActiveMigration == NO) )
    return SYSERR; /* forbidden! */
  ce->importance = htonl(ntohl(ce->importance) + MANAGER_age);
  if ( ntohl(ce->importance) < lowImportance)
    return SYSERR; /* content of this rating was discarded before... */
  if (getSpaceLeft() <= 0) {
    if (sender == NULL)
      LOG(LOG_WARNING,
	  "WARNING: quota for local data store has been reached (%llu kb), can not use more.\n",
	  getDiskQuota());
    return SYSERR; /* drive full */
  }
  /* try to read existing */
  old = NULL;
  oldLen = readContent(dbHandle,
  	 	       &ce->hash,
		       &old);
  /* add the content */
  if (ntohs(ce->type) == LOOKUP_TYPE_3HASH) { 
    /* multiple results are feasible, check! */

    if (oldLen != SYSERR) {
      int ret; 
      char * tmp;

      /* check if the content is already
	 present */
      for (ret=0;ret<oldLen/len;ret++)
	if (0 == memcmp(&(((char*)old)[len*ret]),
			data,
			len)) {
	  /* content already there, abort */
	  xfree(old, 
		"freeing old (db allocated)");
	  *duplicate = YES; 
	  return OK;
	}	   

      tmp = xmalloc(oldLen + len,
		    "allocating buffer for multiple-results");
      memcpy(tmp,
	     old,
	     oldLen);
      memcpy(&tmp[oldLen],
	     data,
	     len);
      ret = writeContent(dbHandle,
      			 &ce->hash,
			 oldLen + len,
			 tmp);
      xfree(tmp, "freeing tmp");
      xfree(old, "freeing old (db allocated)");
      return ret;
    }
  } else {
    if (oldLen == len)
      *duplicate = YES;
  }
  freeNonNull(old, "freeing old (db allocated)");
  if (*duplicate == NO && 
       SYSERR == writeContent(dbHandle,
  		              &ce->hash,
		 	      len,
			      data))
    return SYSERR; /* something went wrong... */
  /* even if it was duplicate, make sure that lookup is consistent */
  if (SYSERR == addEntry(ce, NULL)) {
    if (SYSERR == unlinkFromDB(dbHandle,
    			       &ce->hash)) 
      LOG(LOG_WARNING,
	  "WARNING: double fault, tried to unlink entry from "\
	  "contentdatabase after adding it to the lookup database"\
	  " failed, but this recovery operation also failed.\n");    
    return SYSERR; /* something went wrong */
  }
  return OK;
}


/* end of manager.c */
