/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file include/util/ports.h
 * @author Christian Grothoff
 *
 * Definition of various message numbers (GNUnet port numbers) and
 * some common message types.
 **/

#ifndef PORTS_H
#define PORTS_H

/**
 * we use an unsigned short in the protocol header, thus: 
 **/
#define MAX_BUFFER_SIZE 65536

/* ********* transport protocol numbers ************* */

/**
 * These are the protocol numbers for the underlying GNUnet
 * protocols. They are typically taken to hint at a well-known
 * protocol, but they are not used in the same way. They just indicate
 * *internally+ to GNUnet which protocol from the TCP/IP *suite* to
 * use to run GNUnet over.
 **/
   
/**
 * protocol number for "unspecified"
 **/
#define ANY_PROTOCOL_NUMBER 0

/**
 * protocol number of TCP. Do NEVER change, also used in other context! 
 **/
#define TCP_PROTOCOL_NUMBER 6

/**
 * protocol number of UDP. Do NEVER change, also used in other context! 
 **/
#define UDP_PROTOCOL_NUMBER 17
/**
 * protocol number for SMTP 
 **/
#define SMTP_PROTOCOL_NUMBER 25

/**
 * protocol number for HTTP 
 **/
#define HTTP_PROTOCOL_NUMBER 80


/* ********* client-server protocol (over TCP) ********** */
/* ********* CS infrastructure messages ********** */

/**
 * return value for remote calls 
 **/
#define CS_PROTO_RETURN_VALUE 0

/**
 * client to gnunetd: to how many nodes are we connected? 
 * reply is a CS_RETURN_VALUE message. 
 **/
#define CS_PROTO_CLIENT_COUNT 1

/**
 * Client to gnunetd: how much traffic do we have at the moment?
 **/
#define CS_PROTO_TRAFFIC_QUERY 2

/**
 * gnunetd to client: traffic statistics 
 **/
#define CS_PROTO_TRAFFIC_INFO 3

/* *********** messages for statistics ************* */

/**
 * client to gnunetd: request statistics 
 **/
#define STATS_CS_PROTO_GET_STATISTICS 4

/**
 * gnunetd to client: statistics 
 **/
#define STATS_CS_PROTO_STATISTICS 5

/* ********** CS AFS application messages ********** */

/**
 * client to gnunetd: send queries
 * */
#define AFS_CS_PROTO_QUERY 8

/**
 * gnunetd to client: here is your answer (3-hash-content) 
 **/
#define AFS_CS_PROTO_RESULT_3HASH 9

/**
 * gnunetd to client: here is your answer (CHK-content) 
 **/
#define AFS_CS_PROTO_RESULT_CHK 10

/**
 * client to gnunetd: insert CHK content (no index) 
 **/
#define AFS_CS_PROTO_INSERT_CHK 11

/**
 * client to gnunetd: insert 3HASH content (no index) 
 **/
#define AFS_CS_PROTO_INSERT_3HASH 12

/**
 * client to gnunetd: index content 
 **/
#define AFS_CS_PROTO_INDEX_BLOCK 13

/**
 * client to gnunetd: get an index for a file
 **/
#define AFS_CS_PROTO_INDEX_FILE 14

/**
 * client to gnunetd: index super-block
 **/
#define AFS_CS_PROTO_INDEX_SUPER 15



/* ********** CS CHAT application messages ********** */

#define CHAT_CS_PROTO_MSG 32

/* ********** CS TRACEKIT application messages ********* */

#define TRACEKIT_CS_PROTO_PROBE 36

#define TRACEKIT_CS_PROTO_REPLY 37

/* ********** CS TBENCH application messages ********** */

#define TBENCH_CS_PROTO_REQUEST	40
#define TBENCH_CS_PROTO_REPLY	41

/* **************** common structs ******************* */

/**
 * Header for all Client-Server communications.
 **/
typedef struct {
  /**
   * The length of the struct (in bytes, including the length field itself) 
   **/
  unsigned short size;

  /**
   * The type of the message (XX_CS_PROTO_XXXX) 
   **/
  unsigned short tcpType;

  /**
   * This is followed by a requestType specific data block, consisting
   * of size bytes.
   **/
  char data[0];
} CS_HEADER;

/**
 * CS communication: simple return value
 **/
typedef struct {
  /**
   * The CS header (values: sizeof(CS_RETURN_VALUE), CS_PROTO_RETURN_VALUE)
   **/ 
  CS_HEADER header;

  /**
   * The return value (network byte order) 
   **/
  int return_value;
} CS_RETURN_VALUE;


/* ******** node-to-node (p2p) messages (over anything) ********* */

/* ********* p2p infrastructure messages *********** */

/**
 * announcement of public key 
 **/
#define p2p_PROTO_HELO 0

/**
 * session key exchange, session key is encrypted with hostkey 
 **/
#define p2p_PROTO_SKEY 1

/**
 * PING 
 **/
#define p2p_PROTO_PING 2

/**
 * PONG (response to PING) 
 **/
#define p2p_PROTO_PONG 3

/**
 * timestamp (until when is the packet valid) 
 **/
#define p2p_PROTO_TIMESTAMP 4

/**
 * sequence number (discard packet if sequence number
 * is not higher than previously received number) 
 **/
#define p2p_PROTO_SEQUENCE 5

/**
 * noise, used to fill packets to sizes >1k.
 **/
#define p2p_PROTO_NOISE 6

/**
 * termination of connection (other host is nice
 * and tells us, there is NO requirement to do so!) 
 **/
#define p2p_PROTO_HANGUP 7

/* ************* p2p AFS application messages *********** */

/**
 * Query for content. 
 **/
#define AFS_p2p_PROTO_QUERY 16

/**
 * receive content 
 **/
#define AFS_p2p_PROTO_3HASH_RESULT 17

/**
 * receive CHK content 
 **/
#define AFS_p2p_PROTO_CHK_RESULT 18

/* ************** p2p CHAT application messages *********** */

/**
 * chat message 
 **/
#define CHAT_p2p_PROTO_MSG 32

/* *************** p2p TRACEKIT application messages ******** */

#define TRACEKIT_p2p_PROTO_PROBE 36

#define TRACEKIT_p2p_PROTO_REPLY 37

/* ********** p2p TBENCH application messages ********** */

/**
 * benchmark message: send back reply asap 
 **/
#define TBENCH_p2p_PROTO_REQUEST	40
#define TBENCH_p2p_PROTO_REPLY	41

#define MAX_p2p_PROTO_USED 42

/* ************** common structs ********************* */

/**
 * p2p message part header
 **/
typedef struct {
  /**
   * size of this MESSAGE_PART (network byte order) 
   **/
  unsigned short size;

  /**
   * type of the request, XX_p2p_PROTO_XXX (network byte order) 
   **/
  unsigned short requestType; 

  /**
   * this is followed by a requestType specific data block,
   * consisting of size bytes. 
   **/
  char data[0];
} p2p_HEADER;


#endif
